/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "AMRLevel.H"
#include "AMRLevelPolytropicGas.H"
#include "AMRLevelPolytropicGasFactory.H"


// Flag the use of artificial viscosity and set the coefficient
void AMRLevelPolytropicGasFactory::setArtificialViscosity(bool a_useArtificialViscosity,
                                          Real a_artificialViscosity)
{
  // Artificial viscosity coefficient must be greater than zero
  assert(!a_useArtificialViscosity || (a_artificialViscosity >= 0.0));

  // Store the artificial viscosity flag and coefficient
  m_useArtificialViscosity = a_useArtificialViscosity;
  m_artificialViscosity = a_artificialViscosity;

  m_isArtViscSet = true;
}
/////////
void AMRLevelPolytropicGasFactory::setSlopeParameters(bool a_useFlattening,
                                      bool a_limitSlopes,
                                      bool a_useCharLimiting)
{
  // Store the slope computation parameters
  m_limitSlopes     = a_limitSlopes;
  m_useFlattening   = a_useFlattening;
  m_useCharLimiting = a_useCharLimiting;

  m_isSlopeSet = true;
}
//////////
AMRLevelPolytropicGasFactory::AMRLevelPolytropicGasFactory()
{
  setDefaultValues();
}

// Virtual constructor
AMRLevel* AMRLevelPolytropicGasFactory::new_amrlevel() const
{
  // Make sure everything is defined
  assert(isDefined());

  // Create a new AMRLevelPolytropicGas
  AMRLevelPolytropicGas* amrGodPtr = new AMRLevelPolytropicGas();

  // Set up new object
  amrGodPtr->CFL(m_cfl);
  amrGodPtr->domainLength(m_domainLength);
  amrGodPtr->refinementThreshold(m_refineThresh);
  amrGodPtr->tagBufferSize(m_tagBufferSize);
  amrGodPtr->initialDtMultiplier(m_initialDtMultiplier);
  amrGodPtr->godunovPhysics(m_godunovPhysics);
  amrGodPtr->verbosity(m_verbosity);
  amrGodPtr->setArtificialViscosity(m_useArtificialViscosity,
                                    m_artificialViscosity);
  amrGodPtr->setSlopeParameters(m_useFlattening,
                                m_limitSlopes,
                                m_useCharLimiting);

  // Return it
  return (static_cast <AMRLevel*> (amrGodPtr));
}

AMRLevelPolytropicGasFactory::~AMRLevelPolytropicGasFactory()
{
  if(m_godunovPhysics != NULL)
    delete m_godunovPhysics;
}

// CFL number
void AMRLevelPolytropicGasFactory::CFL(Real a_cfl)
{
  m_cfl = a_cfl;
  m_cflSet = true;
}

void AMRLevelPolytropicGasFactory::verbosity(const int& a_verbosity)
{
  m_verbosity = a_verbosity;
}

// Physical dimension of the longest side of the domain
void AMRLevelPolytropicGasFactory::domainLength(Real a_domainLength)
{
  m_domainLength = a_domainLength;
  m_domainLengthSet = true;
}

// Refinement threshold
void AMRLevelPolytropicGasFactory::refinementThreshold(Real a_refineThresh)
{
  m_refineThresh = a_refineThresh;
  m_refineThreshSet = true;
}

// Tag buffer size
void AMRLevelPolytropicGasFactory::tagBufferSize(int a_tagBufferSize)
{
  m_tagBufferSize = a_tagBufferSize;
  m_tagBufferSizeSet = true;
}

// Initial dt multiplier
void AMRLevelPolytropicGasFactory::initialDtMultiplier(Real a_initialDtMultiplier)
{
  m_initialDtMultiplier = a_initialDtMultiplier;
  m_initialDtMultiplierSet = true;
}

void 
AMRLevelPolytropicGasFactory::godunovPhysics(const GodunovPhysics* const a_godunovPhysics)
{
  if(m_godunovPhysics != NULL)
    delete m_godunovPhysics;
  m_godunovPhysics = a_godunovPhysics->new_godunovPhysics();
  m_godunovSet = true;
}

// Check that everything is defined
bool AMRLevelPolytropicGasFactory::isDefined() const
{
  return (m_cflSet &&
          m_isSlopeSet  &&
          m_domainLengthSet &&
          m_refineThreshSet &&
          m_tagBufferSizeSet &&
          m_initialDtMultiplierSet &&
          m_isArtViscSet &&
          m_godunovSet);
}

// Some default values
void AMRLevelPolytropicGasFactory::setDefaultValues()
{
  CFL(0.8);
  domainLength(1.0);
  refinementThreshold(0.2);
  tagBufferSize(2);
  initialDtMultiplier(0.1);
  m_godunovSet = false;
  m_verbosity = 0;
  m_godunovPhysics = NULL;
}
