/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Mon, July 19, 1999

#ifndef _POISSONBC_H_
#define _POISSONBC_H_

#include "Box.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "LoHiSide.H"
#include "GhostBC.H"

/// Class to enforce homogeneous Neumann boundary conditions 
/**
    BoxGhostBC class implentation to enforce homogeneous neumann 
    boundary conditions 
    Recall that  BoxGhostBC is a class to encapsulate the 
    operations of ghost-cell boundary conditions at a face.  
    If the solution is phi and the face normal direction is x,
    the boundary conditions usually used can be expressed as
    A*phi + B*dphi/dx = C.
    For homogeneous Neumann BCs, A=C=0, B= 1
    These functions are only meant to be called by DomainGhostBC
 */
class NeumannBC: public BoxGhostBC
{
public:
  friend class DomainGhostBC;

  /// Null constructor
  NeumannBC();

  /// Basic constructor -- sets Interval to (0,0)
  NeumannBC(int dir,
            Side::LoHiSide sd);

  /// Full constructor
  NeumannBC(int             dir,
            Side::LoHiSide  sd, 
            const Interval& a_comps);

  /// Destructor
  ~NeumannBC();

protected:
  /// 
  /**
     Fills coefficient arrays -- 
     fills neumann factor with 1.0
     and other two fields with 0.0.
   */
  virtual void fillBCValues(FArrayBox& a_neumfac,
                            FArrayBox& a_dircfac,
                            FArrayBox& a_inhmval,
                            Real       a_dx,
                            const Box& a_domain) const;

  ///
  /**
     Fills coefficient arrays --
     fills neumann factor with 1.0
     and other two fields with 0.0.
   */
  virtual void fillBCValues(FArrayBox&           a_neumfac,
                            FArrayBox&           a_dircfac,
                            FArrayBox&           a_inhmval,
                            Real                 a_dx,
                            const ProblemDomain& a_domain) const;

  ///
  /**
     Virtual constructor workaround.
  */
  BoxGhostBC* new_boxghostbc() const;

private:
  NeumannBC(const NeumannBC& a_neuin){;}
  virtual void operator=(const NeumannBC& a_neuin){;}
};

/// Class to enforce homogeneous Dirichlet boundary conditions 
/**
    BoxGhostBC class implentation to enforce homogeneous dirchelet 
    boundary conditions 
    Recall that
    BoxGhostBC is a class to encapsulate the 
    operations  of ghost-cell boundary conditions at a face.  
    If the solution is phi and the face normal direction is x,
    the boundary conditions usually used can be expressed as
    A*phi + B*dphi/dx = C.
    For homogeneous Dirichlet BCs, B=C=0, A= 1
    These functions are only meant to be called by DomainGhostBC
 */
class DirichletBC: public BoxGhostBC
{
public:
  friend class DomainGhostBC;

  /// Null constructor
  DirichletBC();

  /// Simple constructor (sets Interval to (0,0)
  DirichletBC(int dir, Side::LoHiSide sd);

  /// Full constructor
  DirichletBC(int dir, Side::LoHiSide sd, 
              const Interval& a_comps);

  /// Destructor
  ~DirichletBC();

private:
  ///
  /**
     fills Dirichlet factor with 1.0
     and other two fields with 0.0
   */
  virtual void fillBCValues(FArrayBox& a_neumfac,
                            FArrayBox& a_dircfac,
                            FArrayBox& a_inhmval,
                            Real       a_dx,
                            const Box& a_domain) const;

  ///
  /**
     fills Dirichlet factor with 1.0
     and other two fields with 0.0
   */
  virtual void fillBCValues(FArrayBox&           a_neumfac,
                            FArrayBox&           a_dircfac,
                            FArrayBox&           a_inhmval,
                            Real                 a_dx,
                            const ProblemDomain& a_domain) const;

  ///
  BoxGhostBC* new_boxghostbc() const;

private:
  DirichletBC(const DirichletBC& a_dircin){;}
  virtual void operator=(const DirichletBC& a_dircin){;}
};

#endif
