/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _AMRLEVEL_H_
#define _AMRLEVEL_H_

#include "REAL.H"
#include "Box.H"
#include "DisjointBoxLayout.H"
#include "ProblemDomain.H"

class HDF5Handle;
class IntVectSet;
template <class TTYPE> class Vector;

/// Abstract base class for time-dependent data at a level of refinement.
/**
   This is an abstract base class for data at the same level of
   refinement within a hierarchy of levels.  The concrete class
   derived from AMRLevel is called a  physics class.  The domain
   of a level is a disjoint union of rectangles in a logically
   rectangular index space.  Data is defined within this domain.
   There is also a problem domain, which may be larger, within which
   data can, in theory, be interpolated from some coarser level.

   AMRLevel is intimately entangled with the AMR class.
   The AMR contains a collection of AMRLevel objects which
   represent a hierarchy of levels of refinement.  The class  AMR
   is responsible for calling the correct member functions of
   AMRLevel.  The user is responsible for implementing the physics
   class, and might never call any AMRLevel functions described
   here.
*/

class AMRLevel
{
public:
  ///
  /**
     Default constructor.
  */
  AMRLevel();

  ///
  /**
     Destructor.
  */
  virtual
    ~AMRLevel();

  ///
  /**
     Defines this AMRLevel.

     <ul>
       <li>a_coarser_level_ptr (not modified): pointer to next coarser level
           object.
       <li>a_problem_domain (not modified): problem domain of this level.
       <li>a_level (not modified): index of this level.  The base level is
           zero.
       <li>a_ref_ratio (not modified): the refinement ratio between this level
           and the next finer level.
     </ul>
  */
  virtual
    void define(AMRLevel*  a_coarser_level_ptr,
                const Box& a_problem_domain,
                int        a_level,
                int        a_ref_ratio);

  ///
  /**
     Defines this AMRLevel.

     <ul>
       <li>a_coarser_level_ptr (not modified): pointer to next coarser level
           object.
       <li>a_problem_domain (not modified): problem domain of this level.
       <li>a_level (not modified): index of this level.  The base level is
           zero.
       <li>a_ref_ratio (not modified): the refinement ratio between this level
           and the next finer level.
     </ul>
  */
  virtual
    void define(AMRLevel*            a_coarser_level_ptr,
                const ProblemDomain& a_problem_domain,
                int                  a_level,
                int                  a_ref_ratio);

  ///
  /**
     Sets the pointer-to-finer-level member to a_finer_level_ptr.
  */
  virtual
    void finerLevelPtr(AMRLevel* a_finer_level_ptr);

  ///
  /**
     Sets the time step to a_dt.
  */
  virtual
    void dt(Real a_dt);

  ///
  /**
     Sets the time to a_time.

  */
  virtual
    void time(Real a_time);

  ///
  /**
     Returns the current value of the time step.

  */
  virtual
    Real dt() const;

  ///
  /**
     Returns the current value of the time on this level.
  */
  virtual
    Real time() const;

  ///
  /**
     Sets the initial dt multiplier to a_initial_dt_multiplier.
  */
  virtual
    void initialDtMultiplier(Real a_initial_dt_multiplier);

  ///
  /**
     Returns the initial dt multiplier.
  */
  virtual
    Real initialDtMultiplier() const;

  ///
  /**
     Returns the problem domain of this level.

  */
  virtual
    const ProblemDomain& problemDomain() const;

  ///
  /**
     Returns the domain of this level.

  */
  virtual
    Vector<Box> boxes() const;

  ///
  /**
     Returns true if any AMRLevel::define function has been called,
     false otherwise.
  */
  bool isDefined() const;

  ///
  /**
     Returns the refinement ratio between this level and the next finer level.

  */
  virtual
    int refRatio() const;

  ///
  /**
     Sets verbosity level to a_verbosity.  Minimum verbosity is 0, for
     which nothing is printed.
  */
  static
    void verbosity(int a_verbosity);

  ///
  /**
     Returns current verbosity level.  Minimum verbosity is 0, for
     which nothing is printed.

  */
  static
    int verbosity();

  ///
  /**
     Advances this level by one time step.

     This is a pure virtual function and MUST be defined in the derived
     class.

  */
  virtual
    Real advance() = 0;

  ///
  /**
     Things to do after advancing this level by one time step.

     This is a pure virtual function and  MUST be defined in the derived
     class.

  */
  virtual
    void postTimeStep() = 0;

  ///
  /**
     Creates tagged cells for dynamic mesh refinement.

     This is a pure virtual function and  MUST be defined in the derived
     class.

  */
  virtual
    void tagCells(IntVectSet& a_tags) = 0;

  ///
  /**
     Creates tagged cells for mesh refinement at initialization.

     This is a pure virtual function and MUST be defined in the derived
     class.

  */
  virtual
    void tagCellsInit(IntVectSet& a_tags) = 0;

  ///
  /**
     Redefines this level to have the specified domain a_new_grids.

     This is a pure virtual function and MUST be defined in the derived
     class.

  */
  virtual
    void regrid(const Vector<Box>& a_new_grids) = 0;

  ///
  /**
      Performs any post-regridding operations which are necessary.

      This is not a pure virtual function to preserve compatibility
      with earlier versions of AMRLevel.  The AMRLevel::postRegrid()
      instantiation is a no-op.
  */
  virtual
    void postRegrid(int a_base_level);

  ///
  /**
     Initializes this level to have the specified domain a_new_grids.

     This is a pure virtual function and MUST be defined in the derived
     class.

  */
  virtual
    void initialGrid(const Vector<Box>& a_new_grids) = 0;

  ///
  /**
     Initializes data.

     This is a pure virtual function and  MUST be defined in the derived
     class.

  */
  virtual
    void initialData() = 0;

  ///
  /**
     Things to do after initialization.

     This is a pure virtual function and  MUST be defined in the derived
     class.

  */
  virtual
    void postInitialize() = 0;

  ///
  /**
     Returns maximum stable time step for this level.

     This is a pure virtual function and  MUST be defined in the derived
     class.

  */
  virtual
    Real computeDt() = 0;

  ///
  /**
      Returns maximum stable time step for this level with initial data.

      This is a pure virtual function and  MUST be defined in the derived
      class.

  */
  virtual
    Real computeInitialDt() = 0;

#ifdef HDF5
  ///
  /**
     Writes checkpoint header.

     This is a pure virtual function and  MUST be defined in the derived
     class.

  */
  virtual
    void writeCheckpointHeader (HDF5Handle& a_handle) const = 0;

  ///
  /**
     Write checkpoint data for this level.

     This is a pure virtual function and  MUST be defined in the derived
     class.

  */
  virtual
    void writeCheckpointLevel (HDF5Handle& a_handle) const = 0;

  ///
  /**
     Reads checkpoint header.

     This is a pure virtual function and  MUST be defined in the derived
     class.

  */
  virtual
    void readCheckpointHeader (HDF5Handle& a_handle) = 0;

  ///
  /**
     Reads checkpoint data for this level.

     This is a pure virtual function and MUST be defined in the derived
     class.

  */
  virtual
    void readCheckpointLevel (HDF5Handle& a_handle) = 0;

  ///
  /**
     Writes plot header.

     This is a pure virtual function and MUST be defined in the derived
     class.

  */
  virtual
    void writePlotHeader (HDF5Handle& a_handle) const = 0;

  ///
  /**
     Write plot file for this level.

     This is a pure virtual function and MUST be defined in the derived
     class.

  */
  virtual
    void writePlotLevel (HDF5Handle& a_handle) const = 0;
#endif

protected:
  // the problem domain
  ProblemDomain m_problem_domain;

  //
  Vector<Box> m_level_grids;

  // the level
  int m_level;

  // refinement ratio
  int m_ref_ratio;

  // initial time step multipier
  Real m_initial_dt_multiplier;

  // time step
  Real m_dt;

  // time
  Real m_time;

  // pointer to next coarser level
  AMRLevel* m_coarser_level_ptr;

  // pointer to next finer level
  AMRLevel* m_finer_level_ptr;

  // verbosity level
  static int s_verbosity;

  bool m_isDefined;
};
#endif
