/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _CFStencil_H_
#define _CFStencil_H_

#include <iostream>
#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include "DisjointBoxLayout.H"
#include "Box.H"
#include "LoHiSide.H"
#include "IntVectSet.H"

/// Class to encapsulate coarse-fine information for AMR operations.
/** 
    CFStencil is a wrapper around a bunch of things
    that allow one to do coarse-fine interpolation.
    This class should be considered internal to QuadCFInterp
    and should not be considered part of the Chombo API.
  */
class CFStencil
{

private:

protected:
  //for internal use
  void  setDefaultValues();

  //the following variables are mainly for self-identification

  // direction normal to interface
  int m_direction;                 

  // interface is on high or low side of grid
  Side::LoHiSide m_hiorlo;

  //
  DataIndex m_dataIndex;

  //the following variables are to be used by others

  /* fine intvects which need to be interpolated */
  IntVectSet m_fineIVS;  

  /* coar intvects where slopes can be computed 
      == coarsened version of fiinterp ivs*/
  IntVectSet m_coarIVS;  

  bool m_isDefined;

public:
  /// {\bf Access functions}

  ///
  /** returns true if CFStencil is defined */
  bool isDefined() const;

  ///
  /**
     are there any interpolation points?  \\
     returns false if so.*/
  bool isEmpty() const;

  ///
  /** get fine intvects which need to be interpolated  \\
      This will be empty if isEmpty() returns true*/
  const IntVectSet& getFineIVS() const ;

  ///
  /** get coarse intvects that underly fiinterpivs.
      This will be empty if isEmpty() returns true
      this is the same as getCoarInterpIVS
  */
  const IntVectSet& getCoarIVS() const ;

  ///
  CFStencil& operator= (const CFStencil& cfs_in);

  ///
  CFStencil(const CFStencil& cfs_in);

  /*dummy constructor so i can use this with 
     LocalBoxLayoutData
   */
  CFStencil(const Box& a_b, int a_nComp);

  void define(const Box& a_b, int a_nComp);

  /// 
  CFStencil();

  /// 
  ~CFStencil();

  /// 
  CFStencil(
            const Box& FineDomain, 
            const Box& grid, 
            const DisjointBoxLayout& Levboxes, 
            const DisjointBoxLayout& LevCBoxes, 
            int RefRatio, 
            int Direction, 
            Side::LoHiSide hiorlo);

  /// 
  CFStencil(
            const ProblemDomain& FineDomain,
            const Box& grid, 
            const DisjointBoxLayout& Levboxes, 
            const DisjointBoxLayout& LevCBoxes, 
            int RefRatio, 
            int Direction, 
            Side::LoHiSide hiorlo);

  ///
  void   define(
                const Box& FineDomain, 
                const Box& grid, 
                const DisjointBoxLayout& fine_boxes, 
                const DisjointBoxLayout& coar_boxes, 
                int Refratio,
                int Direction, 
                Side::LoHiSide hiorlo);

  ///
  void   define(
                const ProblemDomain& FineDomain,
                const Box& grid, 
                const DisjointBoxLayout& fine_boxes, 
                const DisjointBoxLayout& coar_boxes, 
                int Refratio,
                int Direction, 
                Side::LoHiSide hiorlo);

  ///
  void clear();

};
#endif
