/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "ProblemDomain.H"
#include "MayDay.H"

using std::istream;
using std::ws;
ShiftIterator::ShiftIterator(const bool* a_isPeriodic)
{
  m_index = 0;
  computeShifts(a_isPeriodic);
}

ShiftIterator::~ShiftIterator()
{
  m_shift_vectors.clear();
}

void
ShiftIterator::computeShifts(const bool* a_isPeriodic)
{
  m_shift_vectors.resize(0);

  // handle no directions are periodic
  if (D_TERM(!a_isPeriodic[0], 
             && !a_isPeriodic[1],
             && !a_isPeriodic[2])) 
    {
      // return with no periodic directions defined
    }
  // now handle if all directions are periodic
  else if (D_TERM(a_isPeriodic[0],
                  && a_isPeriodic[1],
                  && a_isPeriodic[2])) 
    {
      // loop over all possible shifts in this case (including corners)
      D_TERM(
             for (int xVect=-1; xVect<2; xVect++),
             for (int yVect=-1; yVect<2; yVect++),
             for (int zVect=-1; zVect<2; zVect++) )
        {
          IntVect shiftVect(D_DECL(xVect,yVect,zVect));
          // no point in adding the zero vector
          if (shiftVect != IntVect::Zero) 
            {
              m_shift_vectors.push_back(shiftVect);
            }
        } // end loop over possible shift vectors
    }
  // case where some but not all directions are periodic
  // this could probably also swallow the completely periodic case
  else
    {
      // i think this should work 
      // this is designed to filter out non-periodic directions
      D_TERM(int xMult = (a_isPeriodic[0]?1:0);,
             int yMult = (a_isPeriodic[1]?1:0);,
             int zMult = (a_isPeriodic[2]?1:0););
      
      // in non-periodic dirctions, shift vector component should
      // always be 0
      D_TERM(
             for (int xVect=-1*xMult; xVect<1+xMult; xVect++), 
             for (int yVect=-1*yMult; yVect<1+yMult; yVect++), 
             for (int zVect=-1*zMult; zVect<1+zMult; zVect++)  )
        {
          IntVect shiftVect(D_DECL(xVect, yVect, zVect));
          // no point in adding the zero vector
          if (shiftVect != IntVect::Zero) 
            {
              m_shift_vectors.push_back(shiftVect);
                   }
        } // end loop over possible shift vectors
      
      
    } // end case where some directions are periodic

}


ProblemDomain::ProblemDomain(const Box& a_domBox)
  : m_domainBox(a_domBox)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = false;
  }
  m_shiftIt.computeShifts(m_isPeriodic);
} 


ProblemDomain::ProblemDomain(const Box& a_domBox,
                             const bool* a_isPeriodic)
  : m_domainBox(a_domBox)
{
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = a_isPeriodic[dir];
  }
  m_shiftIt.computeShifts(m_isPeriodic);
} 

ProblemDomain::ProblemDomain(const IntVect& small, const IntVect& big)
  : m_domainBox(small, big)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = false;
  } 
  m_shiftIt.computeShifts(m_isPeriodic);
}

ProblemDomain::ProblemDomain(const IntVect& small, const IntVect& big,
                             const bool* a_isPeriodic)
  : m_domainBox(small, big)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = a_isPeriodic[dir];
  } 
  m_shiftIt.computeShifts(m_isPeriodic);
}

ProblemDomain::ProblemDomain(const IntVect& small, const int* vec_len)
  : m_domainBox(small, vec_len)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = false;
  } 
  m_shiftIt.computeShifts(m_isPeriodic);
}


ProblemDomain::ProblemDomain(const IntVect& small, const int* vec_len,
                             const bool* a_isPeriodic)
  : m_domainBox(small, vec_len)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = a_isPeriodic[dir];
  } 
  m_shiftIt.computeShifts(m_isPeriodic);
}




// returns true if box intersects problem domain
bool
ProblemDomain::intersects(const Box& a_box) const
{
  if (isEmpty() || a_box.isEmpty()) return false;
  // this is kinda ugly, but it should work
  return (D_TERM((m_isPeriodic[0] 
                  || (a_box.smallEnd(0)>m_domainBox.smallEnd(0) 
                      || a_box.bigEnd(0)<m_domainBox.bigEnd(0))),
                 || (m_isPeriodic[1] 
                     || (a_box.smallEnd(1)>m_domainBox.smallEnd(1) 
                         || a_box.bigEnd(1)<m_domainBox.bigEnd(1))),
                 || (m_isPeriodic[2]
                     || (a_box.smallEnd(2)>m_domainBox.smallEnd(2)
                         || a_box.bigEnd(2)<m_domainBox.bigEnd(2)))));
}


bool
ProblemDomain::intersectsNotEmpty(const Box& a_box) const
{
  // this is kinda ugly, but it should work
  return (D_TERM((m_isPeriodic[0] 
                  || (a_box.smallEnd(0)>m_domainBox.smallEnd(0) 
                      || a_box.bigEnd(0)<m_domainBox.bigEnd(0))),
                 || (m_isPeriodic[1] 
                     || (a_box.smallEnd(1)>m_domainBox.smallEnd(1) 
                         || a_box.bigEnd(1)<m_domainBox.bigEnd(1))),
                 || (m_isPeriodic[2]
                     || (a_box.smallEnd(2)>m_domainBox.smallEnd(2)
                         || a_box.bigEnd(2)<m_domainBox.bigEnd(2)))));

}


bool
ProblemDomain::intersects(const Box& box1, const Box& box2) const
{

  bool boxesIntersect = false;
  // first do the obvious and easy check
  boxesIntersect = box1.intersects(box2);
  
  // now do the harder check whether periodic images intersect
  if (!boxesIntersect && isPeriodic())
    {
      if (m_domainBox.contains(box1) && m_domainBox.contains(box2)) 
        { 
          // no need to do any periodic checking
        } 
      else
        {
          // loop over periodic shift directions, shift box1, and 
          // check for intersections
          IntVect shiftMult(m_domainBox.size());
          ShiftIterator shiftIt = m_shiftIt;
          for (shiftIt.begin(); shiftIt.ok(); ++shiftIt) 
            {
              Box shiftBox(box1);
              IntVect shiftVect = shiftMult*shiftIt();
              shiftBox.shift(shiftVect);
              if (shiftBox.intersects(box2)) boxesIntersect = true;
            }
        }
    } // end if we need to do periodic check
  return boxesIntersect;
}
              
          


void
ProblemDomain::setPeriodic(int a_dir, bool a_isPeriodic) 
{
  assert (a_dir < SpaceDim);
  m_isPeriodic[a_dir] = a_isPeriodic;
  m_shiftIt.computeShifts(m_isPeriodic);
}






Box
bdryLo(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box bndryBox;
  if (!a_pd.isPeriodic(a_dir))
    {
      bndryBox = bdryLo(a_pd.domainBox(),a_dir,a_len);
    }
  return bndryBox;
}

Box
bdryHi(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box bndryBox;
  if (!a_pd.isPeriodic(a_dir))
    {
      bndryBox = bdryHi(a_pd.domainBox(),a_dir,a_len);
    }
  return bndryBox;
}

Box
adjCellLo(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box adjBox;
  if (!a_pd.isPeriodic(a_dir)) 
    {
      adjBox = adjCellLo(a_pd.domainBox(),a_dir, a_len);
    }
  return adjBox;

}



Box
adjCellHi(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box adjBox;
  if (!a_pd.isPeriodic(a_dir)) 
    {
      adjBox = adjCellHi(a_pd.domainBox(),a_dir, a_len);
    }
  return adjBox;
}



Box
ProblemDomain::operator&(const Box& b) const
{

  Box intersectBox;
  // check for empty -- if either box is empty, return empty intersectBox
  if (!b.isEmpty() && !isEmpty()) 
    {         
      // do this by creating an intersection box which will
      // take the periodic case into account
      intersectBox = m_domainBox;
      if (isPeriodic()) 
        {
          for (int dir=0; dir<SpaceDim; dir++) 
            {
              if (m_isPeriodic[dir])
                {
                  // the idea behind this is that intersect does nothing in
                  // periodic directions
                  intersectBox.setRange(dir, b.smallEnd(dir), b.size(dir));
                }
            }
        }

      intersectBox &= b;            
    } // end if neither box is empty

  
  return intersectBox;
}



ProblemDomain& 
ProblemDomain::refine(int a_refinement_ratio)
{
  m_domainBox.refine(a_refinement_ratio);
  return *this;
}

ProblemDomain
refine(const ProblemDomain& a_probdomain, int a_refinement_ratio)
{
  
  if (a_probdomain.isEmpty()) return a_probdomain;
  Box newdomain = a_probdomain.domainBox();
  newdomain.refine(a_refinement_ratio);
  
  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}

ProblemDomain& 
ProblemDomain::refine(const IntVect& a_refinement_ratio)
{
  m_domainBox.refine(a_refinement_ratio);
  return *this;
}


ProblemDomain
refine(const ProblemDomain& a_probdomain, const IntVect& a_refinement_ratio)
{
  Box newdomain(a_probdomain.domainBox());
  newdomain.refine(a_refinement_ratio);
  
  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}


ProblemDomain& 
ProblemDomain::coarsen(int a_refinement_ratio)
{
  m_domainBox.coarsen(a_refinement_ratio);
  return *this;
}

ProblemDomain
coarsen(const ProblemDomain& a_probdomain, int a_refinement_ratio)
{
  Box newdomain(a_probdomain.domainBox());
  
  newdomain.coarsen(a_refinement_ratio);
  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}


ProblemDomain&
ProblemDomain::coarsen(const IntVect& a_refinement_ratio)
{
  m_domainBox.refine(a_refinement_ratio);
  return *this;
}

ProblemDomain
coarsen(const ProblemDomain& a_probdomain, 
        const IntVect& a_refinement_ratio)
{
  
  Box newdomain(a_probdomain.domainBox());
  newdomain.coarsen(a_refinement_ratio);

  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}


ostream& operator<< (ostream& os, const ProblemDomain& a_probdomain)
{
  os << a_probdomain.domainBox()
     << ", " << D_TERM( a_probdomain.m_isPeriodic[0],
                        << " " << a_probdomain.m_isPeriodic[1],
                        << " " << a_probdomain.m_isPeriodic[2]);

  if (os.fail())
    MayDay::Error("operator<<(ostream&,ProblemDomain&) failed");

  return os;
}

//
// Copied from <Utility.H>
//
#define CH_IGNORE_MAX 100000

istream&
operator>> (istream& is, ProblemDomain& a_probdomain)
{

  is >> a_probdomain.m_domainBox;
  is >> ws;
  char c;
  is >> c;
  is.putback(c);
  if (c == ',')
    {
      is.ignore(CH_IGNORE_MAX, ',');
      for (int dir=0; dir<SpaceDim; dir++) 
        {
          is >> a_probdomain.m_isPeriodic[dir];
        }
    }

  else
    MayDay::Error("operator>>(istream&,ProblemDomain&): expected \',\'");
  
  if (is.fail())
    MayDay::Error("operator>>(istream&,ProblemDomain&) failed");
  
    return is;
}

  
void 
ProblemDomain::dumpOn(ostream& strm) const
{

  strm << "domainBox: ";
  m_domainBox.dumpOn(strm);
  strm << " isPeriodic = " 
       << D_TERM(m_isPeriodic[0], 
                 << " " << m_isPeriodic[1], 
                 << " " << m_isPeriodic[2])
       << '\n';
  
  if (strm.fail())
    MayDay::Error("ProblemDomain::dumpOn(ostream&) failed");

}

void operator &= (Box& a_box, const ProblemDomain& a_probdomain)
{
  a_box = a_probdomain & a_box;
}

Box operator & (const Box& a_box, const ProblemDomain& a_probdomain)
{
  Box returnBox = a_probdomain & a_box;
  return returnBox;
}
