/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// RefCountedPtr.h
// ==================



#ifndef REFCOUNTEDPTR_H
#define REFCOUNTEDPTR_H

#include "MayDay.H"

///  A reference-counting handle class.
/**
       

        This is to be used as a pointer to class T.  
        This will feel and smell just like a built-in pointer except:
        -#  There is no need to call delete on the pointer.
        -#  The default copy constructor and assignment implement ref-counting.
        -#  The user may call isNonUnique to determine if this pointer is
            the only pointer to the data.  This can be used to hide the
            ref-counting behavior of a class.
        -#  Checks for dereference of a null pointer.

This class is completely inlined.

typical usage:

\code

{
  Box b;
  IntVect a;
  //refCount() == 1
  RefCounterPtr<IntVectSet> set(new IntVectSet()); 
  
  // still just one IntVectSet, but refCount()==2
  RefCountedPtr<IntVectSet>  otherSet(set);
  
  // Like a pointer, modifying one modifies the other
  otherSet->define(b);
  (*set)|=a;
  {
	RefCountedPtr<IntVectSet> anotherSet; // null
	anotherSet = otherSet ;  //now all three have refCount()==3
  }//anotherSet out of scope, so deleted.  IntVectSet NOT deleted.
  
  // set.refCount() == 2
  // otherSet.refCount() == 2;
  // otherset == set;

}
// when all RefCountedPtr's for a given object are deleted, the last
// one calls 'delete' on the member pointer.
\endcode

*/
template <class T>
class RefCountedPtr {
public:
  //
  inline RefCountedPtr(T* ptr = 0);
  //
  inline RefCountedPtr(const RefCountedPtr<T>& other);
  //
  inline ~RefCountedPtr();
  /// assignement operator.  copies pointer member
  /** copies pointer member and integer pointer, decreases refcount of rhs member
	  before assignment.  this refcount increased my one. */
  inline const RefCountedPtr<T>& operator =(const RefCountedPtr<T>& rhs);
  /// dereference access operator.  use like a pointer derefence access function.
  inline T* operator ->();
  inline const T* operator ->() const;
  //
  inline bool isNull() const;
  //
  /// pointer dereference.
  inline T& operator *();
  inline const T& operator *() const;
  //
  /// auto conversion to regular const pointer where required.
  inline operator const T* () const;
  
  /// true when refcount is one.
  inline bool isNonUnique() const;

  inline int refCount() const;

protected:
  T* ptr_;
  int* refCount_;
private:
};

template <class T> inline
RefCountedPtr<T>::RefCountedPtr(T* ptr)
  : ptr_(ptr),
    refCount_(0)
{
  if (ptr_)
    {
      refCount_ = new int;
      if (refCount_ == 0)
        MayDay::Error("RefCountedPtr::RefCountedPtr(T* ptr) out of memory");
      *refCount_ = 1;
    }
}

template <class T> inline
RefCountedPtr<T>::RefCountedPtr(const RefCountedPtr<T>& other)
  : ptr_(other.ptr_),
    refCount_(other.refCount_)
{
  if (refCount_ != 0)
    ++(*refCount_);
}

template <class T> inline
RefCountedPtr<T>::~RefCountedPtr() {
  if (refCount_ != 0 && --(*refCount_) == 0)
    {
      delete ptr_;
      ptr_ = 0;
      delete refCount_;
      refCount_ = 0;
    }
}

template <class T> inline
const RefCountedPtr<T>& RefCountedPtr<T>::operator =(const RefCountedPtr<T>& rhs) {
  if (ptr_ != rhs.ptr_)
    {
      if (refCount_ != 0 && --(*refCount_) == 0)
        {
          delete ptr_;
          delete refCount_;
        }
      ptr_ = rhs.ptr_;
      refCount_ = rhs.refCount_;
      if (refCount_ != 0)
        ++(*refCount_);
    }
  return *this;
}

template <class T> inline
bool RefCountedPtr<T>::isNull() const {
  return (ptr_ == 0);
}

template <class T> inline
T* RefCountedPtr<T>::operator ->() {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator ->() on null pointer");
  return ptr_;
}

template <class T> inline
const T* RefCountedPtr<T>::operator ->() const {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator ->() on null pointer");
  return ptr_;
}

template <class T> inline
T& RefCountedPtr<T>::operator *() {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator *() on null pointer");
  return *ptr_;
}

template <class T> inline
const T& RefCountedPtr<T>::operator *() const {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator *() on null pointer");
  return *ptr_;
}

template <class T> inline
RefCountedPtr<T>::operator const T* () const {
  return ptr_;
}

template <class T> inline
bool RefCountedPtr<T>::isNonUnique() const {
  return refCount_ == 0 ? false : *refCount_ != 1;
}

template <class T> inline
int RefCountedPtr<T>::refCount() const {
  return refCount_ == 0 ? 0 : *refCount_;
}

#endif
